@@ -109,8 +109,7 @@ pub(crate) fn handle_sendmsg(
                 // SAFETY: We do not check AF_ALG or AF_NETLINK for access.
             }
             _ if allow_unsupp_socket => {
-                // SAFETY: No pointer-dereference in access check.
-                return unsafe { Ok(request.continue_syscall()) };
+                // SAFETY: We do not check unsupported sockets for access.
             }
             _ => return Err(Errno::EAFNOSUPPORT),
         };
@@ -476,8 +475,7 @@ pub(crate) fn handle_sendmmsg(
                     // SAFETY: We do not check AF_ALG and AF_NETLINK for access.
                 }
                 _ if allow_unsupp_socket => {
-                    // SAFETY: No pointer-dereference in access check.
-                    return unsafe { Ok(request.continue_syscall()) };
+                    // SAFETY: We do not check unsupported sockets for access.
                 }
                 _ => return Err(Errno::EAFNOSUPPORT),
             };
@@ -807,10 +805,10 @@ fn parse_control_messages(
     while offset < cmsg_buf.len() {
         // SAFETY: Ensuring alignment for `cmsghdr` by starting from a u8 pointer.  The
         // `cmsg_buf` is originally a u8 buffer, which may not satisfy the alignment
-        // requirements of `cmsghdr`.  This cast assumes that the buffer provided by
+        // requirements of `cmsghdr`. This cast assumes that the buffer provided by
         // `request.process.read_mem` is correctly aligned for `cmsghdr` structures, which is true if
         // the buffer is initially populated in a manner adhering to the alignment
-        // requirements of `cmsghdr`.  The caller is responsible for ensuring that `offset`
+        // requirements of `cmsghdr`. The caller is responsible for ensuring that `offset`
         // is correctly aligned for `cmsghdr` when accessing the buffer.
         let (cmsg_header, cmsg_len0): (crate::compat::cmsghdr, usize) = if is32 {
             // SAFETY: See the comment above.
@@ -1069,16 +1067,14 @@ fn parse_control_messages(
                 control_tx_times.push(tx_time); // Store the Tx time
             }
             (libc::IPPROTO_IP, libc::IP_TOS) => {
-                // IP_TOS ancillary data is an int.
-                if data_len != size_of::<libc::c_int>() {
+                // IP_TOS ancillary data is a u8.
+                if data_len != size_of::<u8>() {
                     return Err(Errno::EINVAL);
                 }
-                // SAFETY: Exact int-sized payload validated; bounds and alignment guaranteed
-                // by prior checks and cmsg parsing semantics. Single c_int read is safe.
+                // SAFETY: Exact u8-sized payload validated; bounds and alignment guaranteed
+                // by prior checks and cmsg parsing semantics. Single u8 read is safe.
                 #[expect(clippy::cast_ptr_alignment)]
-                let val = unsafe { *(data.as_ptr() as *const libc::c_int) };
-                // See ip_cmd_send() in Linux' net/ipv4/ip_sockglue.c.
-                let val: u8 = val.try_into().or(Err(Errno::EINVAL))?;
+                let val = unsafe { *(data.as_ptr() as *const u8) };
                 control_ipv4_toses.push(val);
             }
             (libc::IPPROTO_IPV6, libc::IPV6_TCLASS) => {
