@@ -1257,8 +1257,15 @@ fn do_open_crypt(path: &CanonicalPath, mut safe_flags: OFlag) -> Result<OwnedFd,
 #[inline]
 fn to_mode(arg: u64, flags: OFlag, ignore: bool) -> Result<Mode, Errno> {
     if flags.contains(OFlag::O_CREAT) || flags.contains(OFlag::O_TMPFILE) {
-        // SAFETY: Reject invalid mode with creation flags.
-        crate::kernel::to_mode(arg)
+        if ignore {
+            // SAFETY: Strip undefined/invalid mode.
+            // Used by open, openat, and creat.
+            Ok(crate::kernel::to_mode(arg))
+        } else {
+            // SAFETY: Reject undefined/invalid mode.
+            // Used by openat2.
+            crate::kernel::to_mode2(arg)
+        }
     } else if !ignore && arg != 0 {
         // SAFETY: mode argument must be zero without creation flags for openat2.
         Err(Errno::EINVAL)
