@@ -19,7 +19,7 @@ use crate::{
     compat::{fstatat64, statx, STATX_BASIC_STATS, STATX_MODE, STATX_TYPE},
     config::{MAGIC_PREFIX, MMAP_MIN_ADDR},
     confine::{scmp_arch_bits, EOWNERDEAD},
-    fs::{is_sidechannel_device, parse_fd, CanonicalPath, FileInfo, FileType, FsFlags},
+    fs::{parse_fd, CanonicalPath, FileInfo, FileType, FsFlags},
     hash::SydHashSet,
     hook::{SysArg, SysFlags, UNotifyEventRequest},
     kernel::{sandbox_path, to_atflags},
@@ -136,11 +136,6 @@ pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
-    // Return EFAULT here for invalid pointers.
-    if req.data.args[4] < *MMAP_MIN_ADDR {
-        return request.fail_syscall(Errno::EFAULT);
-    }
-
     // SAFETY: Reject undefined/invalid flags.
     // const AT_STATX_SYNC_AS_STAT: AtFlags = AtFlags::empty();
     const AT_STATX_FORCE_SYNC: AtFlags = AtFlags::from_bits_retain(0x2000);
@@ -157,6 +152,11 @@ pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
         Err(errno) => return request.fail_syscall(errno),
     };
 
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[4] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let mut flags = SysFlags::empty();
     let mut fsflags = FsFlags::MUST_PATH;
     if atflags.contains(AtFlags::AT_EMPTY_PATH) {
@@ -214,7 +214,7 @@ pub(crate) fn sys_newfstatat(request: UNotifyEventRequest) -> ScmpNotifResp {
     syscall_stat_handler(request, arg, 2, false)
 }
 
-#[allow(clippy::cognitive_complexity)]
+#[expect(clippy::cognitive_complexity)]
 fn syscall_stat_handler(
     request: UNotifyEventRequest,
     arg: SysArg,
@@ -227,8 +227,8 @@ fn syscall_stat_handler(
         let sandbox = request.get_sandbox();
         let is_lock = sandbox.locked_for(req.pid());
 
-        let is_crypt = sandbox.enabled(Capability::CAP_CRYPT);
-        let is_stat = sandbox.enabled(Capability::CAP_STAT);
+        let has_crypt = sandbox.enabled(Capability::CAP_CRYPT);
+        let has_stat = sandbox.enabled(Capability::CAP_STAT);
 
         // Check for chroot.
         if sandbox.is_chroot() {
@@ -239,6 +239,8 @@ fn syscall_stat_handler(
         // If lock is on do not check for magic path.
         let (mut path, magic) = request.read_path(&sandbox, arg, !is_lock)?;
 
+        let restrict_stat_bdev = !sandbox.flags.allow_unsafe_stat_bdev();
+        let restrict_stat_cdev = !sandbox.flags.allow_unsafe_stat_cdev();
         let mut ghost = false;
         if !is_lock && magic {
             drop(sandbox); // release the read-lock.
@@ -306,40 +308,36 @@ fn syscall_stat_handler(
             drop(sandbox); // release the write-lock.
         } else {
             // Handle fstat for files with encryption in progress.
-            let mut crypt_stat = false;
-            if is_crypt && is_fd {
+            #[expect(clippy::disallowed_methods)]
+            if is_fd && has_crypt {
                 // SAFETY: SysArg.path is None asserting dirfd is Some fd!=AT_FDCWD.
-                #[allow(clippy::disallowed_methods)]
                 let fd = path.dir.as_ref().unwrap();
+                // SAFETY: has_crypt asserts crypt_map is Some.
+                let files = request.crypt_map.as_ref().unwrap();
+
                 if let Ok(info) = FileInfo::from_fd(fd) {
-                    #[allow(clippy::disallowed_methods)]
-                    let files = request.crypt_map.as_ref().unwrap();
-                    for (enc_path, map) in
-                        files.read().unwrap_or_else(|err| err.into_inner()).iter()
-                    {
+                    let files = files.0.lock().unwrap_or_else(|err| err.into_inner());
+                    for (enc_path, map) in files.iter() {
                         if info == map.info {
                             // Found underlying encrypted file for the memory fd.
                             // Note, we only ever attempt to encrypt regular files.
-                            path =
-                                CanonicalPath::new(enc_path.clone(), FileType::Reg, arg.fsflags)?;
-                            crypt_stat = true;
+                            path = CanonicalPath::new_crypt(
+                                path.dir.take().unwrap(),
+                                enc_path.clone(),
+                            );
                             break;
                         }
                     }
-                }
+                } // Lock is released here.
             }
 
             // SAFETY:
             // 1. Allow access to fd-only calls.
             // 2. Allow access to files with encryption in progress.
-            // 3. Allow access to /memfd:syd-*. This prefix is internal
+            // 3. Allow access to !memfd:syd-*. This prefix is internal
             //    to Syd and sandbox process cannot create memory file
             //    descriptors with this name prefix.
-            if is_stat
-                && !crypt_stat
-                && arg.path.is_some()
-                && !path.abs().starts_with(b"/memfd:syd-")
-            {
+            if has_stat && arg.path.is_some() && !path.is_syd_memory_fd() {
                 sandbox_path(
                     Some(&request),
                     &sandbox,
@@ -373,7 +371,7 @@ fn syscall_stat_handler(
         let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
         let mut flags = libc::AT_EMPTY_PATH;
 
-        #[allow(clippy::cast_possible_truncation)]
+        #[expect(clippy::cast_possible_truncation)]
         if arg_stat == 4 {
             // statx
 
@@ -409,13 +407,18 @@ fn syscall_stat_handler(
             // update its access and modification times to match the
             // creation time if it is. This prevents timing attacks on
             // block or character devices like /dev/ptmx using stat.
-            if is_sidechannel_device(statx.stx_mode.into()) {
-                statx.stx_atime = statx.stx_ctime;
-                statx.stx_mtime = statx.stx_ctime;
+            if restrict_stat_bdev || restrict_stat_cdev {
+                let filetype = FileType::from(libc::mode_t::from(statx.stx_mode));
+                if (restrict_stat_bdev && filetype.is_block_device())
+                    || (restrict_stat_cdev && filetype.is_char_device())
+                {
+                    statx.stx_atime = statx.stx_ctime;
+                    statx.stx_mtime = statx.stx_ctime;
+                }
             }
 
             // SAFETY: Restore mask, type and mode, see the comment above.
-            #[allow(clippy::cast_possible_truncation)]
+            #[expect(clippy::cast_possible_truncation)]
             if !basic_stx {
                 if (orig_mask & STATX_TYPE) == 0 {
                     statx.stx_mode &= !libc::S_IFMT as u16;
@@ -457,11 +460,16 @@ fn syscall_stat_handler(
             // update its access and modification times to match the
             // creation time if it is. This prevents timing attacks on
             // block or character devices like /dev/ptmx using stat.
-            if is_sidechannel_device(stat.st_mode) {
-                stat.st_atime = stat.st_ctime;
-                stat.st_mtime = stat.st_ctime;
-                stat.st_atime_nsec = stat.st_ctime_nsec;
-                stat.st_mtime_nsec = stat.st_ctime_nsec;
+            if restrict_stat_bdev || restrict_stat_cdev {
+                let filetype = FileType::from(stat.st_mode);
+                if (restrict_stat_bdev && filetype.is_block_device())
+                    || (restrict_stat_cdev && filetype.is_char_device())
+                {
+                    stat.st_atime = stat.st_ctime;
+                    stat.st_mtime = stat.st_ctime;
+                    stat.st_atime_nsec = stat.st_ctime_nsec;
+                    stat.st_mtime_nsec = stat.st_ctime_nsec;
+                }
             }
 
             let addr = req.data.args[arg_stat];
