@@ -3,10 +3,14 @@
 // src/log.rs: Simple logging on standard error using JSON lines
 //
 // Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
+// Based in part upon musl's src/time/__secs_to_tm.c which is:
+//   Copyright © 2005-2020 Rich Felker, et al.
+//   SPDX-License-Identifier: MIT
 //
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
+    fmt,
     io::Write,
     os::fd::{BorrowedFd, RawFd},
     sync::OnceLock,
@@ -490,45 +494,38 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
         if pid.as_raw() != 0 {
             if add_context {
                 if let Ok(cmd) = proc_cmdline(pid) {
-                    msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
-                } else {
-                    msg.insert("cmd".to_string(), Value::Null);
+                    msg.insert("cmd".to_string(), cmd.to_string().into());
                 }
             } else if let Ok(cmd) = proc_comm(pid) {
-                msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
-            } else {
-                msg.insert("cmd".to_string(), Value::Null);
+                msg.insert("cmd".to_string(), cmd.to_string().into());
             }
             if let Ok(dir) = proc_cwd(pid) {
-                msg.insert("cwd".to_string(), Value::String(dir.to_string()));
-            } else {
-                msg.insert("cwd".to_string(), Value::Null);
+                msg.insert("cwd".to_string(), dir.to_string().into());
             }
             if add_context {
                 if let Ok(tty) = proc_tty(pid) {
-                    msg.insert("tty".to_string(), Value::String(tty.to_string()));
-                } else {
-                    msg.insert("tty".to_string(), Value::Null);
+                    msg.insert("tty".to_string(), tty.to_string().into());
                 }
             }
         }
-        msg.insert("pid".to_string(), Value::Number(pid.as_raw().into()));
+        msg.insert("pid".to_string(), pid.as_raw().into());
     }
 
     // Add current user if add_context.
     if add_context {
-        msg.insert(
-            "uid".to_string(),
-            Value::Number(Uid::current().as_raw().into()),
-        );
+        msg.insert("uid".to_string(), Uid::current().as_raw().into());
     }
 
     // Add "syd" = current thread ID.
     let syd = nix::unistd::gettid().as_raw().into();
     msg.insert("syd".to_string(), Value::Number(syd));
 
-    // Add ISO8601 time.
-    msg.insert("time".to_string(), Value::String(format_iso8601(timestamp)));
+    // Add ISO8601 date, fallback to timestamp.
+    if let Ok(date) = format_iso8601(timestamp) {
+        msg.insert("time".to_string(), date.into());
+    } else {
+        msg.insert("time".to_string(), timestamp.into());
+    }
 
     // Reorder req and informational fields for better visibility.
     if let Some(src) = msg.remove("req") {
@@ -537,9 +534,6 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
     if let Some(m) = msg.remove("msg") {
         msg.insert("msg".to_string(), m);
     }
-    if let Some(err) = msg.remove("error") {
-        msg.insert("error".to_string(), err);
-    }
     if let Some(tip) = msg.remove("tip") {
         msg.insert("tip".to_string(), tip);
     }
@@ -677,45 +671,38 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
         if pid.as_raw() != 0 {
             if add_context {
                 if let Ok(cmd) = proc_cmdline(pid) {
-                    msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
-                } else {
-                    msg.insert("cmd".to_string(), Value::Null);
+                    msg.insert("cmd".to_string(), cmd.to_string().into());
                 }
             } else if let Ok(cmd) = proc_comm(pid) {
-                msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
-            } else {
-                msg.insert("cmd".to_string(), Value::Null);
+                msg.insert("cmd".to_string(), cmd.to_string().into());
             }
             if let Ok(dir) = proc_cwd(pid) {
-                msg.insert("cwd".to_string(), Value::String(dir.to_string()));
-            } else {
-                msg.insert("cwd".to_string(), Value::Null);
+                msg.insert("cwd".to_string(), dir.to_string().into());
             }
             if add_context {
                 if let Ok(tty) = proc_tty(pid) {
-                    msg.insert("tty".to_string(), Value::String(tty.to_string()));
-                } else {
-                    msg.insert("tty".to_string(), Value::Null);
+                    msg.insert("tty".to_string(), tty.to_string().into());
                 }
             }
         }
-        msg.insert("pid".to_string(), Value::Number(pid.as_raw().into()));
+        msg.insert("pid".to_string(), pid.as_raw().into());
     }
 
     // Add current user if add_context.
     if add_context {
-        msg.insert(
-            "uid".to_string(),
-            Value::Number(Uid::current().as_raw().into()),
-        );
+        msg.insert("uid".to_string(), Uid::current().as_raw().into());
     }
 
     // Add "syd" = current thread ID.
     let syd = nix::unistd::gettid().as_raw().into();
     msg.insert("syd".to_string(), Value::Number(syd));
 
-    // Add ISO8601 time
-    msg.insert("time".to_string(), Value::String(format_iso8601(timestamp)));
+    // Add ISO8601 date, fallback to timestamp.
+    if let Ok(date) = format_iso8601(timestamp) {
+        msg.insert("time".to_string(), date.into());
+    } else {
+        msg.insert("time".to_string(), timestamp.into());
+    }
 
     // Reorder req and informational fields for better visibility.
     if let Some(src) = msg.remove("req") {
@@ -724,9 +711,6 @@ pub fn log(level: crate::syslog::LogLevel, timestamp: u64, mut msg: Map<String,
     if let Some(m) = msg.remove("msg") {
         msg.insert("msg".to_string(), m);
     }
-    if let Some(err) = msg.remove("error") {
-        msg.insert("error".to_string(), err);
-    }
     if let Some(tip) = msg.remove("tip") {
         msg.insert("tip".to_string(), tip);
     }
@@ -771,32 +755,10 @@ pub fn now() -> u64 {
 /// The format is currently "YYYYMMDDThhmmssZ". The format may change in
 /// the future but it will always remain conformant to the ISO8601
 /// standard.
-#[expect(clippy::arithmetic_side_effects)]
-fn format_iso8601(timestamp: u64) -> String {
-    let timestamp = match libc::time_t::try_from(timestamp) {
-        Ok(t) => t,
-        Err(_) => return "EOVERFLOW".to_string(),
-    };
-    let mut tm = std::mem::MaybeUninit::uninit();
-
-    // SAFETY: gmtime_r returns NULL on error and set errno(3).
-    if unsafe { libc::gmtime_r(&raw const timestamp, tm.as_mut_ptr()) }.is_null() {
-        return Errno::last().to_string();
-    }
-
-    // SAFETY: gmtime_r returned success.
-    let tm = unsafe { tm.assume_init() };
-
-    let year = tm.tm_year + 1900;
-    // tm_mon ranges from 0 to 11.
-    let month = tm.tm_mon + 1;
-    let day = tm.tm_mday;
-    let hour = tm.tm_hour;
-    let minute = tm.tm_min;
-    let second = tm.tm_sec;
-
-    // Format: YYYYMMDDThhmmssZ
-    format!("{year:04}{month:02}{day:02}T{hour:02}{minute:02}{second:02}Z")
+fn format_iso8601(timestamp: u64) -> Result<String, Errno> {
+    let t = i64::try_from(timestamp).or(Err(Errno::EINVAL))?;
+    let t: Tm = t.try_into()?;
+    Ok(t.to_string())
 }
 
 /// Logs an untrusted buffer, escaping it as hex if it contains control characters.
@@ -820,3 +782,212 @@ pub fn contains_ascii_unprintable(buf: &[u8]) -> bool {
 pub fn is_ascii_printable(byte: u8) -> bool {
     (0x20..=0x7e).contains(&byte)
 }
+
+/// Rust port of musl's `struct tm`.
+#[derive(Clone, Copy, Debug, Eq, PartialEq, Hash)]
+pub struct Tm {
+    tm_year: i32, // years since 1900
+    tm_mon: i32,  // 0..11
+    tm_mday: i32, // 1..31
+    tm_wday: i32, // 0..6, Sunday=0
+    tm_yday: i32, // 0..365
+    tm_hour: i32, // 0..23
+    tm_min: i32,  // 0..59
+    tm_sec: i32,  // 0..60 (leap second tolerated like musl)
+}
+
+impl Tm {
+    /// Get the year of the date.
+    pub fn year(&self) -> i32 {
+        self.tm_year.saturating_add(1900)
+    }
+
+    /// Get the month of the date.
+    pub fn month(&self) -> i32 {
+        self.tm_mon.saturating_add(1)
+    }
+
+    /// Get the month day of the date.
+    pub fn day(&self) -> i32 {
+        self.tm_mday
+    }
+
+    /// Get the hour of the date.
+    pub fn hour(&self) -> i32 {
+        self.tm_hour
+    }
+
+    /// Get the minute of the date.
+    pub fn minute(&self) -> i32 {
+        self.tm_min
+    }
+
+    /// Get the second of the date.
+    pub fn second(&self) -> i32 {
+        self.tm_sec
+    }
+}
+
+impl TryFrom<i64> for Tm {
+    type Error = Errno;
+
+    fn try_from(t: i64) -> Result<Self, Errno> {
+        const LEAPOCH: i64 = 951_868_800; // 2000-03-01 00:00:00 UTC
+        const SECS_PER_DAY: i64 = 86_400;
+        const DAYS_PER_400Y: i64 = 146_097;
+        const DAYS_PER_100Y: i64 = 36_524;
+        const DAYS_PER_4Y: i64 = 1_461;
+
+        // Reject time_t values whose year would overflow int.
+        let limit_unit = 31_622_400i64; // 366 days in seconds
+        let low = i64::from(i32::MIN)
+            .checked_mul(limit_unit)
+            .ok_or(Errno::EOVERFLOW)?;
+        let high = i64::from(i32::MAX)
+            .checked_mul(limit_unit)
+            .ok_or(Errno::EOVERFLOW)?;
+        if t < low || t > high {
+            return Err(Errno::EOVERFLOW);
+        }
+
+        let secs = t.checked_sub(LEAPOCH).ok_or(Errno::EOVERFLOW)?;
+        let mut days = secs.checked_div(SECS_PER_DAY).ok_or(Errno::EOVERFLOW)?;
+        let mut remsecs = secs.checked_rem(SECS_PER_DAY).ok_or(Errno::EOVERFLOW)?;
+        if remsecs < 0 {
+            remsecs = remsecs.checked_add(SECS_PER_DAY).ok_or(Errno::EOVERFLOW)?;
+            days = days.checked_sub(1).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        // Weekday: 2000-03-01 was Wednesday (3)
+        let mut wday = (3i64)
+            .checked_add(days)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_rem(7)
+            .ok_or(Errno::EOVERFLOW)?;
+        if wday < 0 {
+            wday = wday.checked_add(7).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        // Gregorian cycles
+        let mut qc_cycles = days.checked_div(DAYS_PER_400Y).ok_or(Errno::EOVERFLOW)?;
+        let mut remdays = days.checked_rem(DAYS_PER_400Y).ok_or(Errno::EOVERFLOW)?;
+        if remdays < 0 {
+            remdays = remdays.checked_add(DAYS_PER_400Y).ok_or(Errno::EOVERFLOW)?;
+            qc_cycles = qc_cycles.checked_sub(1).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        let mut c_cycles = remdays.checked_div(DAYS_PER_100Y).ok_or(Errno::EOVERFLOW)?;
+        if c_cycles == 4 {
+            c_cycles = c_cycles.checked_sub(1).ok_or(Errno::EOVERFLOW)?;
+        }
+        remdays = remdays
+            .checked_sub(
+                c_cycles
+                    .checked_mul(DAYS_PER_100Y)
+                    .ok_or(Errno::EOVERFLOW)?,
+            )
+            .ok_or(Errno::EOVERFLOW)?;
+
+        let mut q_cycles = remdays.checked_div(DAYS_PER_4Y).ok_or(Errno::EOVERFLOW)?;
+        if q_cycles == 25 {
+            q_cycles = q_cycles.checked_sub(1).ok_or(Errno::EOVERFLOW)?;
+        }
+        remdays = remdays
+            .checked_sub(q_cycles.checked_mul(DAYS_PER_4Y).ok_or(Errno::EOVERFLOW)?)
+            .ok_or(Errno::EOVERFLOW)?;
+
+        let mut remyears = remdays.checked_div(365).ok_or(Errno::EOVERFLOW)?;
+        if remyears == 4 {
+            remyears = remyears.checked_sub(1).ok_or(Errno::EOVERFLOW)?;
+        }
+        remdays = remdays
+            .checked_sub(remyears.checked_mul(365).ok_or(Errno::EOVERFLOW)?)
+            .ok_or(Errno::EOVERFLOW)?;
+
+        let leap = remyears == 0 && (q_cycles != 0 || c_cycles == 0);
+        let leap_i = if leap { 1i64 } else { 0i64 };
+
+        let mut yday = remdays
+            .checked_add(59)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_add(leap_i)
+            .ok_or(Errno::EOVERFLOW)?;
+        let yday_lim = 365i64.checked_add(leap_i).ok_or(Errno::EOVERFLOW)?;
+        if yday >= yday_lim {
+            yday = yday.checked_sub(yday_lim).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        let years = remyears
+            .checked_add(4i64.checked_mul(q_cycles).ok_or(Errno::EOVERFLOW)?)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_add(100i64.checked_mul(c_cycles).ok_or(Errno::EOVERFLOW)?)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_add(400i64.checked_mul(qc_cycles).ok_or(Errno::EOVERFLOW)?)
+            .ok_or(Errno::EOVERFLOW)?;
+
+        // Month & day within "Mar..Feb" year layout.
+        let dim: [i64; 12] = [31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31, 29];
+        let mut months = 0i64;
+        let mut rd = remdays;
+        while months < 12 {
+            let d = *dim
+                .get(usize::try_from(months).or(Err(Errno::EOVERFLOW))?)
+                .ok_or(Errno::EOVERFLOW)?;
+            if d > rd {
+                break;
+            }
+            rd = rd.checked_sub(d).ok_or(Errno::EOVERFLOW)?;
+            months = months.checked_add(1).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        let mut years_adj = years;
+        let mut months_adj = months;
+        if months_adj >= 10 {
+            months_adj = months_adj.checked_sub(12).ok_or(Errno::EOVERFLOW)?;
+            years_adj = years_adj.checked_add(1).ok_or(Errno::EOVERFLOW)?;
+        }
+
+        let years_plus_100 = years_adj.checked_add(100).ok_or(Errno::EOVERFLOW)?;
+        if years_plus_100 > i64::from(i32::MAX) || years_plus_100 < i64::from(i32::MIN) {
+            return Err(Errno::EOVERFLOW);
+        }
+
+        let hour = remsecs.checked_div(3600).ok_or(Errno::EOVERFLOW)?;
+        let min = remsecs
+            .checked_div(60)
+            .ok_or(Errno::EOVERFLOW)?
+            .checked_rem(60)
+            .ok_or(Errno::EOVERFLOW)?;
+        let sec = remsecs.checked_rem(60).ok_or(Errno::EOVERFLOW)?;
+
+        Ok(Self {
+            tm_year: i32::try_from(years_plus_100).or(Err(Errno::EOVERFLOW))?,
+            tm_mon: i32::try_from(months_adj.checked_add(2).ok_or(Errno::EOVERFLOW)?)
+                .or(Err(Errno::EOVERFLOW))?,
+            tm_mday: i32::try_from(rd.checked_add(1).ok_or(Errno::EOVERFLOW)?)
+                .or(Err(Errno::EOVERFLOW))?,
+            tm_wday: i32::try_from(wday).or(Err(Errno::EOVERFLOW))?,
+            tm_yday: i32::try_from(yday).or(Err(Errno::EOVERFLOW))?,
+            tm_hour: i32::try_from(hour).or(Err(Errno::EOVERFLOW))?,
+            tm_min: i32::try_from(min).or(Err(Errno::EOVERFLOW))?,
+            tm_sec: i32::try_from(sec).or(Err(Errno::EOVERFLOW))?,
+        })
+    }
+}
+
+impl fmt::Display for Tm {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        let year = self.year();
+        let month = self.month();
+        let day = self.day();
+        let hour = self.hour();
+        let minute = self.minute();
+        let second = self.second();
+
+        // Format: YYYYMMDDThhmmssZ
+        write!(
+            f,
+            "{year:04}{month:02}{day:02}T{hour:02}{minute:02}{second:02}Z"
+        )
+    }
+}
