@@ -14,7 +14,7 @@ use std::{
     cmp::Ordering,
     collections::VecDeque,
     ffi::{CStr, OsStr, OsString},
-    ops::Deref,
+    ops::{Deref, RangeBounds},
     os::{
         fd::RawFd,
         unix::ffi::{OsStrExt, OsStringExt},
@@ -1205,6 +1205,13 @@ impl XPathBuf {
         Ok(())
     }
 
+    /// Remove the given range of bytes from the path buffer, dropping them.
+    ///
+    /// Panics on out-of-bounds.
+    pub fn drain<R: RangeBounds<usize>>(&mut self, range: R) {
+        self.0.drain(range);
+    }
+
     /// Create a path from the given PID.
     pub fn from_pid(pid: Pid) -> Self {
         let mut buf = itoa::Buffer::new();
