@@ -42,12 +42,12 @@ use nix::{
 use crate::{
     compat::{pipe2_raw, waitid, WaitStatus},
     config::{MINI_STACK_SIZE, *},
-    confine::{confine_mdwe, MS_NOSYMFOLLOW},
+    confine::{confine_mdwe, CLONE_NEWTIME, MS_NOSYMFOLLOW},
     debug,
     err::err2no,
     fs::{
-        create_memfd, safe_clone, safe_copy_if_exists, seal_memfd, set_cloexec, set_nonblock,
-        MFD_ALLOW_SEALING,
+        create_memfd, pidfd_send_signal, safe_clone, safe_copy_if_exists, seal_memfd, set_cloexec,
+        set_nonblock, MFD_ALLOW_SEALING,
     },
     get_user_home, get_user_name,
     hash::SydHashSet,
@@ -307,7 +307,7 @@ impl WordExp<'_> {
         // Create a memory fd to write input into,
         // and pass to the internal /bin/sh invoked
         // by wordexp(3).
-        let mut file = create_memfd(b"syd-exp\0", MFD_ALLOW_SEALING).map(File::from)?;
+        let mut file = create_memfd(b"syd-wordexp\0", MFD_ALLOW_SEALING).map(File::from)?;
         debug!("ctx": "expand",
             "msg": format!("created memory-file {} with close-on-exec flag set",
                 file.as_raw_fd()));
@@ -321,16 +321,15 @@ impl WordExp<'_> {
         file.write_all(b"\n")?;
 
         // Handle user-specific configuration.
-        let home = env::var_os("HOME")
-            .map(XPathBuf::from)
-            .unwrap_or(XPathBuf::from("/proc/self/fdinfo"));
-        safe_copy_if_exists(&mut file, &home.join(b".config/syd/init.sh"))?;
-        file.write_all(b"\n")?;
+        if let Some(home) = env::var_os("HOME").map(XPathBuf::from) {
+            safe_copy_if_exists(&mut file, &home.join(b".config/syd/init.sh"))?;
+            file.write_all(b"\n")?;
+        }
 
         // Write input.
-        file.write_all(b"echo -n ")?;
+        file.write_all(b"eval set -- x ")?;
         file.write_all(input.as_bytes())?;
-        file.write_all(b"\n")?;
+        file.write_all(b"\nshift\nprintf '%s ' \"$@\"\n")?;
 
         // Seal memfd for future writes.
         seal_memfd(&file)?;
@@ -344,7 +343,7 @@ impl WordExp<'_> {
             "msg": format!("set close-on-exec flag to off for memory-file {}",
                 file.as_raw_fd()));
 
-        let shell = format!("`. /proc/self/fd/{}`", file.as_raw_fd());
+        let shell = format!("`. /proc/thread-self/fd/{}`", file.as_raw_fd());
         debug!("ctx": "expand",
             "msg": format!("passing memory file {} to wordexp(3) with {} seconds timeout...",
                 file.as_raw_fd(), timeout.as_secs()));
@@ -440,20 +439,11 @@ impl WordExp<'_> {
 
         loop {
             if !sig && (err as i32 != 0 || epoch.elapsed() >= timeout) {
-                // a. out of memory condition
-                // b. timeout exceeded
-                // send SIGKILL once, and fall-through to wait.
+                // a. Out of memory condition
+                // b. Timeout exceeded
+                // Send SIGKILL once, and fall-through to wait.
                 sig = true;
-                // SAFETY: There's no libc wrapper for pidfd_send_signal yet.
-                let _ = unsafe {
-                    libc::syscall(
-                        libc::SYS_pidfd_send_signal,
-                        pid_fd.as_fd().as_raw_fd(),
-                        SIGKILL,
-                        0,
-                        0,
-                    )
-                };
+                let _ = pidfd_send_signal(&pid_fd, SIGKILL);
             } else if !eof {
                 // read one batch from pipe.
                 match pipe.read(&mut buf) {
@@ -585,12 +575,13 @@ impl WordExp<'_> {
                 | CloneFlags::CLONE_NEWNET
                 | CloneFlags::CLONE_NEWNS
                 | CloneFlags::CLONE_NEWPID
-                | CloneFlags::CLONE_NEWUTS,
+                | CloneFlags::CLONE_NEWUTS
+                | CLONE_NEWTIME,
         )
         .is_ok()
         {
             debug!("ctx": "expand",
-                "msg": "created/entered into new user, mount, pid, network, cgroup, ipc, and uts namespaces");
+                "msg": "created and entered into new user, mount, pid, network, cgroup, ipc, uts, and time namespaces");
             // SAFETY: map current user in new user namespace.
             let uid_buf = {
                 let uid_maps = vec![
@@ -856,7 +847,11 @@ impl WordExp<'_> {
         // Use a binary tree sorted by syscall number, if possible.
         let _ = filter.set_ctl_optimize(2);
 
-        for sysname in WORDEXP_SYSCALLS {
+        for sysname in WORDEXP_SYSCALLS
+            .iter()
+            .chain(GET_ID_SYSCALLS)
+            .chain(VDSO_SYSCALLS)
+        {
             if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
                 filter
                     .add_rule(ScmpAction::Allow, syscall)
@@ -896,11 +891,11 @@ impl WordExp<'_> {
                     "/proc",
                     Some("proc"),
                     flags,
-                    Some("hidepid=2"),
+                    Some("hidepid=4"),
                 ) {
                     Ok(_) => {
                         debug!("ctx": "expand",
-                            "msg": "mounted proc with hidepid=2 in the new mount namespace");
+                            "msg": "mounted proc with hidepid=4 in the new mount namespace");
                     }
                     Err(errno) => {
                         debug!("ctx": "expand",
@@ -938,13 +933,8 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "breakpoint", // arm
     "brk",
     "cacheflush", // arm
-    "cachestat",
     "capget",
     "chdir",
-    "clock_getres",
-    "clock_getres_time64",
-    "clock_gettime",
-    "clock_gettime64",
     "clock_nanosleep",
     "clock_nanosleep_time64",
     "clone",
@@ -992,16 +982,7 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "get_mempolicy",
     "get_robust_list",
     "get_thread_area",
-    "getcpu",
     "getcwd",
-    "getegid",
-    "getegid32",
-    "geteuid",
-    "geteuid32",
-    "getgid",
-    "getgid32",
-    "getgroups",
-    "getgroups32",
     "getitimer",
     "getpeername",
     "getpgid",
@@ -1010,17 +991,11 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "getpmsg",
     "getppid",
     "getpriority",
-    "getrandom",
-    "getresgid",
-    "getresuid",
     "getrlimit",
     "getrusage",
     "getsid",
     "getsockopt",
     "gettid",
-    "gettimeofday",
-    "getuid",
-    "getuid32",
     "getxattr",
     "io_cancel",
     "io_destroy",
@@ -1162,7 +1137,6 @@ const WORDEXP_SYSCALLS: &[&str] = &[
     "sync_file_range",
     "tee",
     "tgkill",
-    "time",
     "timer_create",
     "timer_delete",
     "timer_getoverrun",
