@@ -26,7 +26,7 @@ use libseccomp::{scmp_cmp, RawSyscall, ScmpAction, ScmpFilterContext, ScmpSyscal
 use nix::{
     errno::Errno,
     fcntl::{posix_fadvise, splice, tee, OFlag, PosixFadviseAdvice, SpliceFFlags},
-    unistd::{lseek64, pipe2, write, Gid, Uid, Whence},
+    unistd::{lseek64, write, Gid, Uid, Whence},
 };
 use serde::{Serialize, Serializer};
 
@@ -39,9 +39,10 @@ use crate::{
         scmp_add_setid_rules, ExportMode,
     },
     cookie::{
-        safe_ftruncate64, ACCEPT4_COOKIE_ARG4, ACCEPT4_COOKIE_ARG5, FTRUNCATE64_COOKIE_ARG4,
-        FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2, FTRUNCATE_COOKIE_ARG3,
-        FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, SYS_ACCEPT4,
+        safe_ftruncate64, safe_pipe2, ACCEPT4_COOKIE_ARG4, ACCEPT4_COOKIE_ARG5,
+        FTRUNCATE64_COOKIE_ARG4, FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2,
+        FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, PIPE2_COOKIE_ARG2,
+        PIPE2_COOKIE_ARG3, PIPE2_COOKIE_ARG4, PIPE2_COOKIE_ARG5, SYS_ACCEPT4,
     },
     debug,
     err::{err2no, SydJoinHandle, SydResult},
@@ -154,7 +155,7 @@ impl AesWorker {
     }
 
     /// Confine AES thread.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
         flags: Flags,
         transit_uids: &[(Uid, Uid)],
@@ -168,8 +169,8 @@ impl AesWorker {
         ctx.set_ctl_nnp(true)?;
 
         // Disable Speculative Store Bypass mitigations
-        // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
+        // with trace/allow_unsafe_exec_speculative:1
+        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
 
         // DO NOT synchronize filter to all threads.
         // Main thread will confine itself.
@@ -249,7 +250,7 @@ impl AesWorker {
 
         // ftruncate{,64}(2) may be used only with syscall argument cookies.
         let sysname = "ftruncate";
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::useless_conversion)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 // Secure using syscall argument cookies.
@@ -276,7 +277,7 @@ impl AesWorker {
         }
 
         let sysname = "ftruncate64";
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::useless_conversion)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 // Secure using syscall argument cookies.
@@ -306,10 +307,8 @@ impl AesWorker {
         //
         // We only enforce this on architectures where the system call is direct,
         // and there's no socketcall(2) multiplexer indirection.
-        #[allow(clippy::cast_possible_truncation)]
-        #[allow(clippy::cast_sign_loss)]
-        #[allow(clippy::useless_conversion)]
-        #[allow(deprecated)]
+        #[expect(clippy::cast_possible_truncation)]
+        #[expect(clippy::useless_conversion)]
         if let Some(syscall) = SYS_ACCEPT4.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
             // Secure using syscall argument cookies.
             let mut rules = vec![];
@@ -338,19 +337,23 @@ impl AesWorker {
             }
         }
 
-        // Deny pipe2(2) O_NOTIFICATION_PIPE flag.
+        // Restrict pipe2(2) flags.
+        // Restrict pipe2(2) using syscall argument cookies.
         let sysname = "pipe2";
-        #[allow(clippy::cast_sign_loss)]
+        #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::useless_conversion)]
         if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
-            // O_NOTIFICATION_PIPE is equivalent to O_EXCL,
-            // see: linux/watch_queue.h
-            const O_NOTIFICATION_PIPE: u64 = OFlag::O_EXCL.bits() as u64;
+            let mut rules = vec![scmp_cmp!($arg1 == OFlag::O_CLOEXEC.bits() as u64)];
+            if !flags.allow_unsafe_nocookie() {
+                rules.extend(&[
+                    scmp_cmp!($arg2 == (*PIPE2_COOKIE_ARG2).into()),
+                    scmp_cmp!($arg3 == (*PIPE2_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*PIPE2_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*PIPE2_COOKIE_ARG5).into()),
+                ]);
+            }
 
-            ctx.add_rule_conditional(
-                ScmpAction::Allow,
-                syscall,
-                &[scmp_cmp!($arg1 & O_NOTIFICATION_PIPE == 0)],
-            )?;
+            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
         } else {
             info!("ctx": "confine", "op": "allow_aes_syscall",
                 "msg": format!("invalid or unsupported syscall {sysname}"));
@@ -372,6 +375,17 @@ impl AesWorker {
             }
         }
 
+        // Allow safe prctl(2) operations.
+        let sysname = "prctl";
+        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+            for (_, op) in AES_PRCTL_OPS {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
+            }
+        } else {
+            info!("ctx": "confine", "op": "allow_aes_syscall",
+                "msg": format!("invalid or unsupported syscall {sysname}"));
+        }
+
         // Deny installing new signal handlers for {rt_,}sigaction(2).
         for sysname in ["sigaction", "rt_sigaction"] {
             let syscall = match ScmpSyscall::from_name(sysname) {
@@ -389,7 +403,7 @@ impl AesWorker {
         }
 
         // Allow safe system calls.
-        for sysname in AES_SYSCALLS {
+        for sysname in AES_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
@@ -444,7 +458,7 @@ impl AesWorker {
         Ok(ctx)
     }
 
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn try_spawn(self) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_aes".to_string())
@@ -560,7 +574,7 @@ impl AesWorker {
                 .stack_size(AES_STACK_SIZE)
                 .spawn(move || {
                     // SAFETY: crypt_map keys are valid FDs.
-                    #[allow(unsafe_code)]
+                    #[expect(unsafe_code)]
                     let crypt_fd = unsafe { OwnedFd::from_raw_fd(crypt_data.crypt_fd) };
 
                     // Wait until we take a write lock on the encrypted fd.
@@ -597,7 +611,7 @@ impl AesWorker {
         Ok(handle)
     }
 
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     fn sync<Fd: AsFd>(
         fdalg: (RawFd, RawFd),
         crypt_fd: Fd,
@@ -617,11 +631,11 @@ impl AesWorker {
         let file_mode = crypt_data.mode;
         // syd_aes thread steals the IV,
         // therefore it is always Some.
-        #[allow(clippy::disallowed_methods)]
+        #[expect(clippy::disallowed_methods)]
         let mut iv = crypt_data.iv.unwrap();
 
         // SAFETY: crypt_data.enc_fd is a valid FD.
-        #[allow(unsafe_code)]
+        #[expect(unsafe_code)]
         let enc_fd = unsafe { OwnedFd::from_raw_fd(crypt_data.enc_fd) };
         debug!("ctx": "aes", "op": "encrypt_start",
             "path": crypt_path, "memfd": memfd, "mode": file_mode,
@@ -661,7 +675,7 @@ impl AesWorker {
         }
 
         // Handle opened for append but appended nothing quickly.
-        #[allow(clippy::arithmetic_side_effects)]
+        #[expect(clippy::arithmetic_side_effects)]
         if is_append
             && data_size
                 <= file_size.saturating_sub((CRYPT_MAGIC.len() + HMAC_TAG_SIZE + IV_SIZE) as u64)
@@ -683,7 +697,7 @@ impl AesWorker {
         let sock_mac = hmac_sha256_init(&mac_fd, false)?;
         hmac_sha256_feed(&sock_mac, CRYPT_MAGIC, true)?;
         hmac_sha256_feed(&sock_mac, iv.as_ref(), true)?;
-        let (pipe_rd_mac, pipe_wr_mac) = pipe2(OFlag::O_CLOEXEC)?;
+        let (pipe_rd_mac, pipe_wr_mac) = safe_pipe2(OFlag::O_CLOEXEC)?;
 
         // Hint the kernel about I/O access patterns.
         Self::advise_io(&crypt_fd, &enc_fd)?;
@@ -756,7 +770,7 @@ impl AesWorker {
         // The IV is no longer needed.
         drop(iv);
 
-        let (pipe_rd_enc, pipe_wr_enc) = pipe2(OFlag::O_CLOEXEC)?;
+        let (pipe_rd_enc, pipe_wr_enc) = safe_pipe2(OFlag::O_CLOEXEC)?;
 
         let options = AesWorkerOpts {
             memfd,
@@ -856,7 +870,7 @@ impl AesWorker {
             "path": crypt_path, "memfd": memfd, "mode": file_mode,
             "msg": "feed ciphertext into HMAC algoritm socket");
 
-        #[allow(clippy::cast_possible_wrap)]
+        #[expect(clippy::cast_possible_wrap)]
         lseek64(&enc_fd, SYD3_HDR_SIZE as i64, Whence::SeekSet)?;
 
         let mut remain: usize = last_block_offset.try_into().or(Err(Errno::EOVERFLOW))?;
@@ -932,7 +946,7 @@ impl AesWorker {
         let buf = &CRYPT_MAGIC;
         let mut nwrite = 0;
         while nwrite < buf.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match write(&enc_fd, &buf[nwrite..]) {
                 Ok(0) => return Err(Errno::EINVAL),
                 Ok(n) => nwrite += n,
@@ -950,7 +964,7 @@ impl AesWorker {
         let hmac_placeholder = [0u8; HMAC_TAG_SIZE];
         let mut nwrite = 0;
         while nwrite < hmac_placeholder.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match write(&enc_fd, &hmac_placeholder[nwrite..]) {
                 Ok(0) => return Err(Errno::EINVAL),
                 Ok(n) => nwrite += n,
@@ -963,7 +977,7 @@ impl AesWorker {
         let buf = iv.as_ref();
         let mut nwrite = 0;
         while nwrite < buf.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match write(&enc_fd, &buf[nwrite..]) {
                 Ok(0) => return Err(Errno::EINVAL),
                 Ok(n) => nwrite += n,
@@ -1075,7 +1089,6 @@ impl AesWorker {
     }
 
     /// Write HMAC tag to the encrypted file.
-    #[allow(clippy::cognitive_complexity)]
     fn hmac_write<Fd1: AsFd, Fd2: AsFd>(
         enc_fd: Fd1,
         sock_mac: Fd2,
@@ -1093,7 +1106,7 @@ impl AesWorker {
         let hmac_tag = hmac_sha256_fini(&sock_mac)?;
 
         // Seek back to the position after the magic header.
-        #[allow(clippy::cast_possible_wrap)]
+        #[expect(clippy::cast_possible_wrap)]
         lseek64(&enc_fd, CRYPT_MAGIC.len() as i64, Whence::SeekSet)?;
 
         // Write the HMAC tag to the file.
@@ -1103,7 +1116,7 @@ impl AesWorker {
         let buf = hmac_tag.as_slice();
         let mut nwrite = 0;
         while nwrite < buf.len() {
-            #[allow(clippy::arithmetic_side_effects)]
+            #[expect(clippy::arithmetic_side_effects)]
             match write(&enc_fd, &buf[nwrite..]) {
                 Ok(0) => return Err(Errno::EINVAL),
                 Ok(n) => nwrite += n,
