@@ -28,9 +28,8 @@ use crate::{
     alert,
     cache::SysInterrupt,
     config::*,
-    confine::{scmp_add_setid_rules, ExportMode},
+    confine::{confine_scmp_write, scmp_add_setid_rules, ExportMode},
     cookie::{OPENAT2_COOKIE_ARG4, OPENAT2_COOKIE_ARG5},
-    crit,
     err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
     fs::{closeexcept, retry_on_eintr, seccomp_notify_id_valid, tgkill},
@@ -99,29 +98,12 @@ impl Interrupter {
         }
     }
 
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn try_spawn(self, notif_pipe: (RawFd, RawFd)) -> Result<SydJoinHandle<()>, Errno> {
         thread::Builder::new()
             .name("syd_int".to_string())
             .stack_size(INT_STACK_SIZE)
             .spawn(move || {
-                // SAFETY: Default panic hook wont work with stdio(3) closed.
-                std::panic::set_hook(Box::new(|info| {
-                    let err = match info.payload().downcast_ref::<&'static str>() {
-                        Some(s) => *s,
-                        None => match info.payload().downcast_ref::<String>() {
-                            Some(s) => &**s,
-                            None => "?",
-                        },
-                    };
-
-                    let file = info.location().map(|l| l.file());
-                    let line = info.location().map(|l| l.line());
-
-                    crit!("ctx": "panic", "op": "panic_interrupt_thread",
-                        "msg": err, "file": file, "line": line);
-                }));
-
                 // SAFETY: We use exit_group(2) here to bail,
                 // because this unsharing is a critical safety feature.
                 if let Err(errno) = unshare(CloneFlags::CLONE_FS | CloneFlags::CLONE_FILES) {
@@ -140,7 +122,7 @@ impl Interrupter {
                 };
                 drop(pipe_rd);
                 let buf = [42u8; 1];
-                #[allow(clippy::disallowed_methods)]
+                #[expect(clippy::disallowed_methods)]
                 match retry_on_eintr(|| write(&pipe_wr, &buf)).unwrap() {
                     0 => return Err(Errno::EIO.into()), // Syd died before reading.
                     1 => {}
@@ -155,7 +137,7 @@ impl Interrupter {
                 // 2. Static FD of procfs(5).
                 // 3. Log FD.
                 // We have to sort the set as the FDs are randomized.
-                #[allow(clippy::cast_sign_loss)]
+                #[expect(clippy::cast_sign_loss)]
                 let mut set = vec![
                     self.seccomp_fd as libc::c_uint,
                     PROC_FD() as libc::c_uint,
@@ -264,7 +246,7 @@ impl Interrupter {
         //    process is no longer valid otherwise we may end up with a
         //    deadlock: See miniupnpc tests, thx kepstin!
         // 2. To prevent PID reuse vectors we validate the request ID.
-        #[allow(clippy::cast_possible_wrap)]
+        #[expect(clippy::cast_possible_wrap)]
         let pid = Pid::from_raw(interrupt.request.pid as libc::pid_t);
         let status = match proc_status(pid) {
             Ok(status) if self.is_valid(interrupt.request.id) => {
@@ -336,7 +318,7 @@ impl Interrupter {
     }
 
     // Interrupt the respective `syd_emu` thread.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     fn interrupt(tgid: Pid, interrupt: &SysInterrupt, source: InterruptSource) {
         match tgkill(tgid, interrupt.handler, libc::SIGALRM) {
             Ok(_) | Err(Errno::ESRCH) => {
@@ -362,7 +344,7 @@ impl Interrupter {
     }
 
     /// Confine Interrupter thread.
-    #[allow(clippy::cognitive_complexity)]
+    #[expect(clippy::cognitive_complexity)]
     pub(crate) fn prepare_confine(
         seccomp_fd: RawFd,
         tgid: Pid,
@@ -384,8 +366,8 @@ impl Interrupter {
         ctx.set_ctl_nnp(true)?;
 
         // Disable Speculative Store Bypass mitigations
-        // with trace/allow_unsafe_spec_exec:1
-        ctx.set_ctl_ssb(flags.allow_unsafe_spec_exec())?;
+        // with trace/allow_unsafe_exec_speculative:1
+        ctx.set_ctl_ssb(flags.allow_unsafe_exec_speculative())?;
 
         // DO NOT synchronize filter to all threads.
         // Other threads will self-confine.
@@ -405,7 +387,7 @@ impl Interrupter {
         // Allow interrupt handler thread to send the
         // SIGALRM signal to threads in Syd's thread group.
         let sysname = "tgkill";
-        #[allow(clippy::cast_sign_loss)]
+        #[expect(clippy::cast_sign_loss)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -426,8 +408,8 @@ impl Interrupter {
         // Allow interrupt handler thread to
         // validate seccomp request IDs using ioctl(2).
         let sysname = "ioctl";
-        #[allow(clippy::cast_sign_loss)]
-        #[allow(clippy::unnecessary_cast)]
+        #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::unnecessary_cast)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -448,8 +430,8 @@ impl Interrupter {
         // Allow openat2 with the static procfs(5) fd only.
         // Apply system call argument cookies.
         let sysname = "openat2";
-        #[allow(clippy::cast_sign_loss)]
-        #[allow(clippy::useless_conversion)]
+        #[expect(clippy::cast_sign_loss)]
+        #[expect(clippy::useless_conversion)]
         match ScmpSyscall::from_name(sysname) {
             Ok(syscall) => {
                 ctx.add_rule_conditional(
@@ -500,6 +482,17 @@ impl Interrupter {
             }
         }
 
+        // Allow safe prctl(2) operations.
+        let sysname = "prctl";
+        if let Ok(syscall) = ScmpSyscall::from_name(sysname) {
+            for (_, op) in INT_PRCTL_OPS {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg0 == *op)])?;
+            }
+        } else {
+            info!("ctx": "confine", "op": "allow_int_syscall",
+                "msg": format!("invalid or unsupported syscall {sysname}"));
+        }
+
         // Prevent executable memory.
         const PROT_EXEC: u64 = libc::PROT_EXEC as u64;
         for sysname in ["mmap", "mmap2", "mprotect"] {
@@ -518,36 +511,12 @@ impl Interrupter {
             }
         }
 
-        // SAFETY: Allow write(2) to syd::log::LOG_FD only.
-        // If logging is disabled, deny write(2) completely.
-        // Note, we do allow the use-case where LOG_FD is
-        // intentionally set to a negative value to disable
-        // logging, hence we cast directly to an u64 and not
-        // use u64::try_from!
-        // Exception: Allow write(2) globally if profiling is enabled.
-        let log_fd = crate::log::LOG_FD.load(Ordering::Relaxed);
-        for sysname in ["write", "writev"] {
-            #[allow(clippy::cast_sign_loss)]
-            match ScmpSyscall::from_name(sysname) {
-                Ok(syscall) if cfg!(feature = "prof") => {
-                    ctx.add_rule(ScmpAction::Allow, syscall)?;
-                }
-                Ok(syscall) => {
-                    ctx.add_rule_conditional(
-                        ScmpAction::Allow,
-                        syscall,
-                        &[scmp_cmp!($arg0 == log_fd as u64)],
-                    )?;
-                }
-                Err(_) => {
-                    info!("ctx": "confine", "op": "allow_int_syscall",
-                        "msg": format!("invalid or unsupported syscall {sysname}"));
-                }
-            }
-        }
+        // Allow writes to the log-fd.
+        // No proc_pid_mem(5) access required here.
+        confine_scmp_write(&mut ctx, None, false)?;
 
         // Allow safe system calls.
-        for sysname in INT_SYSCALLS {
+        for sysname in INT_SYSCALLS.iter().chain(VDSO_SYSCALLS) {
             match ScmpSyscall::from_name(sysname) {
                 Ok(syscall) => {
                     ctx.add_rule(ScmpAction::Allow, syscall)?;
